import os
import subprocess
import tempfile
import threading
import time
from enum import IntEnum
from typing import Union

import miniaudio

import deepmine.speech.v1.speech_pb2 as speech_pb2
from main.exceptions import KnownAudioConversionError, UnknownAudioConversionError


class AudioEncoding(IntEnum):
    ENCODING_UNSPECIFIED = speech_pb2.RecognitionConfig.AudioEncoding.ENCODING_UNSPECIFIED
    LINEAR16 = speech_pb2.RecognitionConfig.AudioEncoding.LINEAR16
    FLAC = speech_pb2.RecognitionConfig.AudioEncoding.FLAC
    MULAW = speech_pb2.RecognitionConfig.AudioEncoding.MULAW
    OGG_OPUS = speech_pb2.RecognitionConfig.AudioEncoding.OGG_OPUS


# https://github.com/devsnd/tinytag/blob/master/tinytag/tinytag.py
class StreamingContent(miniaudio.StreamableSource):
    def __init__(self, buffer_size=128 * 1024):
        # print('Initializing the stream')
        self._index = 0
        self._buffer = b''
        self._finished = False
        self._buffer_size = buffer_size
        self._buffer_lock = threading.Lock()

    def write(self, segment: bytes):
        while 0 < len(self._buffer) > self._buffer_size:
            time.sleep(0.01)
            if self._finished:
                return
        with self._buffer_lock:
            self._buffer += segment
            # print(f'Adding segment with len {len(segment)}, buffer len is {len(self._buffer)}')

    def read(self, num_bytes: int) -> Union[bytes, memoryview]:
        # print(f'Request for read {num_bytes} bytes.')
        while not self._finished and len(self._buffer) - self._index < num_bytes:
            time.sleep(0.01)
        with self._buffer_lock:
            segment = self._buffer[self._index:self._index + num_bytes]
            self._index += len(segment)
            if self._index > 0.7 * self._buffer_size:
                # print('Truncate the buffer')
                self._buffer = self._buffer[self._index:]
                self._index = 0
            # print(f'Read a segment with len {len(segment)}, current index is {self._index}')
            return segment

    def seek(self, offset: int, origin: miniaudio.SeekOrigin) -> bool:
        # print(f'Seek request to offset {offset} and with origin {origin}')
        with self._buffer_lock:
            if origin == miniaudio.SeekOrigin.START and offset <= len(self._buffer):
                self._index = offset
                # print(f'Seek index to {self._index}')
                return True
            elif origin == miniaudio.SeekOrigin.CURRENT and offset <= len(self._buffer) - self._index:
                self._index += offset
                # print(f'Seek index to {self._index}')
                return True
        return False

    def close(self) -> None:
        # print('Finishing the stream.')
        self._finished = True

    @property
    def is_finished(self):
        return self._finished


class Converter(object):

    @staticmethod
    def audio_file_to_pipe(filename: str, output_sample_rate: int) -> bytes:
        try:
            result = miniaudio.decode_file(filename=filename, output_format=miniaudio.SampleFormat.SIGNED16,
                                           nchannels=1, sample_rate=output_sample_rate,
                                           dither=miniaudio.DitherMode.NONE)
            return result.samples.tobytes()
        except miniaudio.DecodeError:
            print('Using ffmpeg')
            process = subprocess.Popen(['ffmpeg', '-loglevel', 'quiet', '-i', filename, '-ar', str(output_sample_rate),
                                        '-ac', '1', '-f', 's16le', '-'], stdin=None, stdout=subprocess.PIPE)

            return process.stdout.read()

    @staticmethod
    def known_audio_buffer_to_pipe(content: bytes, output_sample_rate: int) -> bytes:
        try:
            result = miniaudio.decode(data=content, output_format=miniaudio.SampleFormat.SIGNED16, nchannels=1,
                                      sample_rate=output_sample_rate, dither=miniaudio.DitherMode.NONE)
            return result.samples.tobytes()
        except miniaudio.MiniaudioError as exp:
            print(exp)
            raise KnownAudioConversionError()

    # https://stackoverflow.com/questions/57958480/ffmpeg-cant-stream-aac-files-from-stdin
    @staticmethod
    def audio_buffer_to_pipe(content: bytes, output_sample_rate: int) -> bytes:
        try:
            process = subprocess.Popen(['ffmpeg', '-loglevel', 'quiet', '-i', 'pipe:0', '-ar', str(output_sample_rate),
                                        '-ac', '1', '-f', 's16le', '-'], shell=False, stdin=subprocess.PIPE,
                                       stdout=subprocess.PIPE)
            result = process.communicate(content)[0]
            if len(result) > 0:
                return result
            # For some situation such as AAC in MPEG4 container, converting from pipe does not work.
            # Save the content to a temp file and try again from file
            temp = tempfile.mkstemp()[1]
            with open(temp, 'wb') as _fid:
                _fid.write(content)
            process = subprocess.Popen(['ffmpeg', '-loglevel', 'quiet', '-i', temp, '-ar', str(output_sample_rate),
                                        '-ac', '1', '-f', 's16le', '-'], stdin=None, stdout=subprocess.PIPE)
            result = process.stdout.read()
            if os.path.exists(temp):
                try:
                    os.remove(temp)
                except IOError:
                    pass
            return result
        except Exception as exp:
            print(exp)
            raise UnknownAudioConversionError()

    @staticmethod
    def conversion_stream(content: StreamingContent, source_format: AudioEncoding, output_sample_rate: int):
        if source_format == AudioEncoding.LINEAR16 or source_format == AudioEncoding.MULAW:
            _format = miniaudio.FileFormat.WAV
        elif source_format == AudioEncoding.FLAC:
            _format = miniaudio.FileFormat.FLAC
        elif source_format == AudioEncoding.OGG_OPUS:
            _format = miniaudio.FileFormat.VORBIS
        else:
            _format = miniaudio.FileFormat.UNKNOWN
        stream = miniaudio.stream_any(source=content, source_format=_format,
                                      output_format=miniaudio.SampleFormat.SIGNED16, nchannels=1,
                                      sample_rate=output_sample_rate)
        return stream


def _read_content(file_path: str, streaming_content: StreamingContent):
    print('Starting get_content')
    with open(file_path, 'rb') as _fid:
        # if file_path.endswith('.wav'):
        #     _fid.read(44)
        while True:
            segment = _fid.read(2048)
            if not segment:
                break
            streaming_content.write(segment)
            # print('Read content with len {}'.format(len(segment)))
            # time.sleep(.1)
            # break
    streaming_content.close()


if __name__ == '__main__':
    # converted = audio_file_to_pipe(r"C:\Users\Hossein\Downloads\Music\LDC2011S08.wav", 16000)
    # with open(r"samples\music.m4a", 'rb') as fid:
    # with open(r"C:\Users\Hossein\Downloads\Telegram Desktop\convertedAudio.3gp", 'rb') as fid:
    #     converted = Converter.audio_buffer_to_pipe(fid.read(), 16000)
    # with open('converted.wav', 'wb') as fid:
    #     fid.write(converted)
    # main()
    _streaming_content = StreamingContent()
    # thread_pool = futures.ThreadPoolExecutor(2)
    # future = thread_pool.submit(_read_content, ("samples/music.flac", _streaming_content))
    # print(future.done())
    thread = threading.Thread(target=_read_content, args=("samples/music.flac", _streaming_content))
    thread.daemon = True
    thread.start()
    time.sleep(2)
    segment_iterator = Converter.conversion_stream(_streaming_content, AudioEncoding.FLAC, 16000)
    print(type(segment_iterator))
    # time.sleep(5)
    print('Start for loop')
    with open('converted.wav', 'wb') as fid:
        for _segment in segment_iterator:
            print(type(_segment))
            fid.write(_segment.tobytes())
    # print(future.done())
    #
    # url = "http://ice5.somafm.com/groovesalad-128-mp3"
    #
    # def title_printer(client: miniaudio.IceCastClient, new_title: str) -> None:
    #     print("Stream title: ", new_title)
    #
    # source = miniaudio.IceCastClient(url, update_stream_title=title_printer)
    # print("Connected to internet stream, audio format:", source.audio_format.name)
    # print("Station name: ", source.station_name)
    # print("Station genre: ", source.station_genre)
    # print("Press <enter> to quit playing.\n")
    # stream = miniaudio.stream_any(source, miniaudio.FileFormat.UNKNOWN)# source.audio_format)
    # with miniaudio.PlaybackDevice() as device:
    #     device.start(stream)
    #     input()
    # source.close()
