import json

from main import status
from main.translations import translations


class GrpcException(Exception):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    default_detail = 'A server error occurred.'
    default_code = 'error'

    @classmethod
    def detail(cls, status_code=None, detail=None, code=None, lang='fa'):
        if status_code is None:
            status_code = cls.status_code
        if detail is None:
            detail = cls.default_detail
            if lang in translations:
                trans = translations[lang].get(cls.default_code)
                if trans:
                    detail = trans
        if code is None:
            code = cls.default_code
        return json.dumps({
            'code': code,
            'status': status_code,
            'message': detail
        })


class NoAccessToken(GrpcException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = 'The access-token is required to be sent via metadata header.'
    default_code = 'token_not_provided'


class NoUserId(GrpcException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = 'The user-id is required to be sent via metadata header by api admin.'
    default_code = 'user_id_not_provided'


class InvalidAdminToken(GrpcException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = 'The provided admin token is not valid.'
    default_code = 'invalid_admin_token'


class InvalidToken(GrpcException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = 'Token is invalid or expired'
    default_code = 'token_not_valid'


class InvalidTokenFormat(GrpcException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = 'The provided token is not a JWT token or not issued by DeepMine!'
    default_code = 'token_format_not_valid'


class NoRemainingQuotaError(GrpcException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'There is not enough remaining quota to transcribe the input speech.'
    default_code = 'no_remaining_quota'


class NoRemainingNonbasicQuotaError(GrpcException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'There is not enough remaining nonbasic quota to transcribe the input speech in streaming mode ' \
                     'or convert long uploaded files.'
    default_code = 'no_remaining_nonbasic_quota'


class KnownAudioConversionError(GrpcException):
    status_code = status.HTTP_422_UNPROCESSABLE_ENTITY
    default_detail = 'A problem was happened during known audio conversion. Maybe the audio format is not supported.'
    default_code = 'known_audio_conversion_error'


class UnknownAudioConversionError(GrpcException):
    status_code = status.HTTP_422_UNPROCESSABLE_ENTITY
    default_detail = 'A problem was happened during unknown audio conversion. Maybe the audio format is not supported.'
    default_code = 'unknown_audio_conversion_error'


class AudioContentError(GrpcException):
    status_code = status.HTTP_422_UNPROCESSABLE_ENTITY
    default_detail = 'The audio content is empty. Maybe the audio format is not supported and an error was happened ' \
                     'during format conversion.'
    default_code = 'audio_content_error'


class InvalidUriFormatError(GrpcException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'The format of provided URI is not valid. Currently we only support http/https/ftp uris. '
    default_code = 'invalid_uri_format_error'


class UriDownloadError(GrpcException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'Failed to download audio from the provided uri. May the URI does not exist. '
    default_code = 'uri_download_error'


class InvalidFirstStreamingRequest(GrpcException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'The first `StreamingRecognizeRequest` message must contain a ' \
                     '`streaming_config` message and must not contain `audio_content`.'
    default_code = 'invalid_first_streaming_request'


class InvalidStreamHeader(GrpcException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'Unable to detect audio format details from uploaded content header. Note that if conversion ' \
                     'will be needed, the audio header should be provided. You can provide the audio in LINEAR16 ' \
                     'with sampling frequency 16K (or 8K for telephony models) to by pass conversion in serverside.'
    default_code = 'invalid_stream_header'


class InvalidSubsequentStreamingRequest(GrpcException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'All subsequent "StreamingRecognizeRequest" messages must contain ' \
                     '`audio_content` and must not contain a `streaming_config` message.'
    default_code = 'invalid_subsequent_streaming_request'


class UnimplementedError(GrpcException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'The requested feature is not supported yet.'
    default_code = 'unimplemented_error'


class InvalidArgumentError(GrpcException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'The provided argument is wrong or is not supported.'
    default_code = 'invalid_argument_error'


class QuotaServerTimeoutError(GrpcException):
    status_code = status.HTTP_408_REQUEST_TIMEOUT
    default_detail = 'Unable to connect to the quota server, please try again later.'
    default_code = 'quota_server_timeout_error'


class HTTPError(GrpcException):
    default_code = 'http_error'

    def __init__(self, status_code, reason=None, api_message=None):
        if 400 <= status_code < 500:
            self._code = 'http_client_error'
        elif 500 <= status_code < 600:
            self._code = 'http_server_error'
        else:
            self._code = HTTPError.default_code
        if isinstance(reason, bytes):
            try:
                reason = reason.decode('utf-8')
            except UnicodeDecodeError:
                reason = reason.decode('iso-8859-1')
        self._reason = reason
        self._status_code = status_code
        self._api_message = api_message if '"detail":' in api_message else None

    def __str__(self):
        return json.dumps({
            'code': self._code,
            'status': self._status_code,
            'message': self._reason,
            'api_message': self._api_message
        })
