import threading
import time

import jwt

from main.exceptions import InvalidTokenFormat, QuotaServerTimeoutError
from main.rest_api_client import api_get_remaining_quota_by_user_id, api_get_remaining_quota, api_use_quota
from settings import SEC_BETWEEN_TWO_UPDATE, IS_DEVELOPMENT, logger


class UserQuota(object):

    def __init__(self, user_id):
        self.user_id = user_id
        self._lock = threading.Lock()
        self._remaining_quota = [0, 0]
        self._unsent_quota = [0, 0]
        self._last_fetch_at = 0

    def update_remaining_quota(self, access_token=None, user_id=None):
        with self._lock:
            if user_id:
                remaining_quota = api_get_remaining_quota_by_user_id(user_id)
            elif access_token:
                remaining_quota = api_get_remaining_quota(access_token)
            else:
                raise Exception("One of access-token or user-id should be un-none")
            if remaining_quota is None:
                logger.warn(f'Request timeout in update_remaining_quota for user: '
                            f'{user_id if user_id else access_token[:36]}')
                if sum(self._remaining_quota) < 300:
                    raise QuotaServerTimeoutError("Unable to connect to the quota server, please try again later.")
                return
            basic_remaining_quota = remaining_quota['basic_remaining_quota_sec']
            nonbasic_remaining_quota = remaining_quota['nonbasic_remaining_quota_sec']
            if self._last_fetch_at == 0:
                self._remaining_quota = [basic_remaining_quota, nonbasic_remaining_quota]
                self._last_fetch_at = time.time()
                return
            temp = abs(basic_remaining_quota - self._unsent_quota[0] - self._remaining_quota[0])
            if temp > 0.1:
                logger.warn('Invalid basic remaining quota, local: {:.2f}, server: {:d}, and unsent: {:.2f}'
                            .format(self._remaining_quota[0], basic_remaining_quota, self._unsent_quota[0]))
                self._remaining_quota[0] = basic_remaining_quota - self._unsent_quota[0]
                self._last_fetch_at = time.time()
            temp = abs(nonbasic_remaining_quota - self._unsent_quota[1] - self._remaining_quota[1])
            if temp > 0.1:
                logger.warn('Invalid nonbasic remaining quota, local: {:.2f}, server: {:d}, and unsent: {:.2f}'
                            .format(self._remaining_quota[1], nonbasic_remaining_quota, self._unsent_quota[1]))
                self._remaining_quota[1] = nonbasic_remaining_quota - self._unsent_quota[1]
                self._last_fetch_at = time.time()

    def get_remaining_quota(self, nonbasic: bool) -> int:
        with self._lock:
            if nonbasic:
                return self._remaining_quota[1]
            else:
                return sum(self._remaining_quota)

    def use_quota(self, duration_sec, nonbasic: bool, client_ip: str, for_roundup=False, is_admin_call=False):
        if duration_sec <= 0:
            return
        with self._lock:
            if nonbasic:
                if self._remaining_quota[1] < duration_sec:
                    if not for_roundup:
                        logger.warn('Duration {:.2f} is more than nonbasic remaining {:.2f}'
                                    .format(duration_sec, self._remaining_quota[1]))
                    duration_sec = self._remaining_quota[1]
                self._remaining_quota[1] -= duration_sec
                self._unsent_quota[1] += duration_sec
                if self._unsent_quota[1] >= SEC_BETWEEN_TWO_UPDATE:
                    self._send_usage_quota(client_ip, is_admin_call=is_admin_call)
                return self._remaining_quota[1]

            _sum = sum(self._remaining_quota)
            if _sum < duration_sec:
                if not for_roundup:
                    logger.warn('Duration {:.2f} is more than total remaining {:.2f}'
                                .format(duration_sec, _sum))
                duration_sec = _sum

            if duration_sec <= self._remaining_quota[1]:
                self._remaining_quota[1] -= duration_sec
                self._unsent_quota[1] += duration_sec
            else:
                duration_sec -= self._remaining_quota[1]
                self._remaining_quota[1] = 0
                self._unsent_quota[1] += self._remaining_quota[1]
                self._remaining_quota[0] -= duration_sec
                self._unsent_quota[0] += duration_sec

            if sum(self._unsent_quota) > SEC_BETWEEN_TWO_UPDATE:
                self._send_usage_quota(client_ip, is_admin_call=is_admin_call)
            return sum(self._remaining_quota)

    def _send_usage_quota(self, client_ip, finalize=False, is_admin_call=False):
        if finalize:
            basic_used_sec = int(self._unsent_quota[0])
            nonbasic_used_sec = int(self._unsent_quota[1])
            if basic_used_sec + nonbasic_used_sec >= 2:
                remaining_quota = api_use_quota(self.user_id, basic_used_sec, nonbasic_used_sec)
                if remaining_quota is None:
                    logger.warn(f'Request timeout in send_usage_quota for user: {self.user_id}')
                    if basic_used_sec + nonbasic_used_sec > 600:
                        raise QuotaServerTimeoutError("Unable to connect to the quota server, please try again later.")
                    return
                if IS_DEVELOPMENT or is_admin_call:
                    logger.info(f'uid: {self.user_id:<5} ip: {client_ip:<15} Remaining quota, basic: '
                                f'{remaining_quota["basic_remaining_quota_sec"]} , nonbasic: '
                                f'{remaining_quota["nonbasic_remaining_quota_sec"]}')
                self._unsent_quota[0] -= basic_used_sec
                self._unsent_quota[1] -= nonbasic_used_sec
        else:
            basic_used_sec = int(self._unsent_quota[0] / 10)
            nonbasic_used_sec = int(self._unsent_quota[1] / 10)
            if basic_used_sec + nonbasic_used_sec > 0:
                remaining_quota = api_use_quota(self.user_id, basic_used_sec * 10, nonbasic_used_sec * 10)
                if remaining_quota is None:
                    logger.warn(f'Request timeout in send_usage_quota for user: {self.user_id}')
                    if basic_used_sec + nonbasic_used_sec > 60:
                        raise QuotaServerTimeoutError("Unable to connect to the quota server, please try again later.")
                    return
                if IS_DEVELOPMENT or is_admin_call:
                    logger.info(f'uid: {self.user_id:<5} ip: {client_ip:<15} Remaining quota, basic: '
                                f'{remaining_quota["basic_remaining_quota_sec"]} , nonbasic: '
                                f'{remaining_quota["nonbasic_remaining_quota_sec"]}')
                self._unsent_quota[0] %= 10
                self._unsent_quota[1] %= 10

    def finalize_usage(self, client_ip):
        with self._lock:
            self._send_usage_quota(client_ip, finalize=True)

    def __str__(self):
        return str('user_id: %s' % self.user_id)


class QuotaManager(object):
    users = {}
    _lock = threading.Lock()

    @classmethod
    def get_user_quota_class(cls, access_token: str = None, user_id: str = None) -> UserQuota:
        if not user_id:
            try:
                encoded_access_token = jwt.decode(access_token, algorithms=["HS256"],
                                                  options={'verify_signature': False})
            except jwt.exceptions.PyJWTError:
                raise InvalidTokenFormat()
            user_id = encoded_access_token.get('user_id')
            if user_id is None:
                raise InvalidTokenFormat()
        with cls._lock:
            user_quota = cls.users.get(user_id)
            if user_quota is not None:
                return user_quota
            user_quota = UserQuota(user_id)
            cls.users[user_id] = user_quota
            return user_quota
