#!/usr/bin/env python3
import array
import json
import logging
import os
import random
import string
import threading
import time
from collections import defaultdict
from concurrent import futures
from copy import deepcopy
from enum import IntEnum

import grpc
import miniaudio
import validators
from vosk import Model, KaldiRecognizer, SetLogLevel

import deepmine.speech.v1.speech_pb2 as speech_pb2
import deepmine.speech.v1.speech_pb2_grpc as speech_pb2_grpc
# Uncomment for better memory usage
# import gc
# gc.set_threshold(0)
from main.audio_utils import StreamingContent, AudioEncoding, Converter
from main.exceptions import NoAccessToken, NoRemainingQuotaError, InvalidSubsequentStreamingRequest, \
    InvalidFirstStreamingRequest, NoRemainingNonbasicQuotaError, InvalidStreamHeader, NoUserId, InvalidAdminToken, \
    GrpcException, AudioContentError, InvalidUriFormatError, UriDownloadError, InvalidArgumentError, UnimplementedError
from main.quota import QuotaManager
from main.utils import levenshtein, download_file
from settings import AVAILABLE_MODELS, MIN_QUOTA_SEC, TIMEOUT_SEC, DURATION_LIMIT_SEC, MAX_MESSAGE_LENGTH, \
    ASR_SERVER_PORT, ASR_SERVER_INTERFACE, API_ADMIN_TOKEN, MAX_WORKERS, MAXIMUM_CONCURRENT_RPCS, IS_DEVELOPMENT, \
    SAVE_AUDIOS, logger

SetLogLevel(0)


class SpeechEventType(IntEnum):
    SPEECH_EVENT_UNSPECIFIED = speech_pb2.SpeechEventType.SPEECH_EVENT_UNSPECIFIED
    END_OF_SINGLE_UTTERANCE = speech_pb2.SpeechEventType.END_OF_SINGLE_UTTERANCE
    END_POINT = speech_pb2.SpeechEventType.END_POINT
    END_OF_SPEECH = speech_pb2.SpeechEventType.END_OF_SPEECH
    DATA_SIZE_LIMIT = speech_pb2.SpeechEventType.DATA_SIZE_LIMIT
    TIMEOUT = speech_pb2.SpeechEventType.TIMEOUT
    QUOTA_FINISHED = speech_pb2.SpeechEventType.QUOTA_FINISHED
    CANCELLED = 12


class Aligner(object):

    def __init__(self, text):
        self.text = text
        self.overall_results = []

    @staticmethod
    def get_word_info(json_word):
        # speaker_tag is always zero now
        return speech_pb2.WordInfo(start_time=json_word['start'], end_time=json_word['end'],
                                   word=json_word['word'], confidence=json_word['conf'])

    def add_result(self, json_result):
        result = json.loads(json_result)
        text = result['text'].strip()
        if len(text) == 0:
            return
        print(text)
        for json_word in result.get('result', []):
            self.overall_results.append(self.get_word_info(json_word))

    def do_alignment(self):
        original_words = self.text.split()
        recognized_words = [word.word for word in self.overall_results]
        edits, cost = levenshtein(original_words, recognized_words)
        results = []
        idx = 0
        _len = len(edits)
        inserted = []
        deleted = []
        while idx < _len:
            t, i, j = edits[idx]
            idx += 1
            if len(deleted) > 0 and (t == 'm' or t == 's'):
                word = self.overall_results[j]
                cnt = 1
                char_cnt = len(original_words[i])
                while cnt <= len(deleted):
                    if deleted[-cnt] not in word.word and cnt > 1:
                        break
                    char_cnt += len(deleted[-cnt])
                    cnt += 1
                new_words = []
                confidence = word.confidence if t == 'm' else max(word.confidence * (1.0 - 0.1 * cnt), 0.5)
                char_duration = (word.end_time - word.start_time) / char_cnt
                last_start = word.end_time - char_duration * len(original_words[i])
                new_words.append(speech_pb2.WordInfo(start_time=last_start, end_time=word.end_time,
                                                     word=original_words[i], confidence=confidence))
                if cnt > 1:
                    while cnt > 1 and len(deleted) > 0:
                        new_word = deleted.pop()
                        print('d1 %s' % new_word)
                        end_time = last_start
                        confidence = 0.1 * word.confidence
                        last_start = end_time - char_duration * len(new_word)
                        new_words.append(speech_pb2.WordInfo(start_time=last_start, end_time=end_time,
                                                             word=new_word, confidence=confidence))
                        cnt -= 1
                start_time = 0 if len(results) == 0 else results[-1].end_time
                for new_word in reversed(deleted):
                    print('d2 %s' % new_word)
                    new_words.append(speech_pb2.WordInfo(start_time=start_time, end_time=start_time,
                                                         word=new_word, confidence=0.05))

                for word in reversed(new_words):
                    print(
                        'ms %s - [ %.3f - %.3f ]  %.2f' % (word.word, word.start_time, word.end_time, word.confidence))
                    results.append(word)
                deleted.clear()
                inserted.clear()
            elif t == 'm':  # match
                inserted.clear()
                deleted.clear()
                word = self.overall_results[j]
                print('m  %s - [ %.3f - %.3f ]  %.2f' % (word.word, word.start_time, word.end_time, word.confidence))
                results.append(self.overall_results[j])
            elif t == 's':  # substitution
                word = deepcopy(self.overall_results[j])
                with_inserted = ' '
                if len(inserted) > 0:
                    confidence = word.confidence
                    start = word.start_time
                    orig_word = original_words[i]
                    cnt = 1
                    for _word in reversed(inserted):
                        if _word.word not in orig_word and cnt > 1:
                            break
                        start = _word.start_time
                        confidence += _word.confidence
                        cnt += 1
                    word.word = orig_word
                    word.confidence = max(confidence / cnt * (1.0 - 0.1 * cnt), 0.5)
                    word.start_time = start
                    inserted.clear()
                    with_inserted = 'i'
                else:
                    word.word = original_words[i]
                    word.confidence = word.confidence * 0.85
                print('s%s %s - [ %.3f - %.3f ]  %.2f' % (with_inserted, word.word, word.start_time,
                                                          word.end_time, word.confidence))
                results.append(word)
            elif t == 'd':  # deletion
                deleted.append(original_words[i])
            elif t == 'i':
                inserted.append(self.overall_results[j])

        if len(results) == 0:
            speech_pb2.AlignmentResponse(words=results, confidence=0)

        confidence = 0
        for word in results:
            confidence += word.confidence
        return speech_pb2.AlignmentResponse(words=results, confidence=confidence / len(results))


class ResponseGenerator(object):

    def __init__(self, non_streaming=False, enable_word_confidence=False, enable_word_time_offsets=False, lm_order=3):
        self.lm_order = lm_order
        self.last_partial_result = ''
        self.non_streaming = non_streaming
        self.results_buffer = []
        self.enable_word_confidence = enable_word_confidence
        self.enable_word_time_offsets = enable_word_time_offsets

    @staticmethod
    def get_word_info(json_word):
        # speaker_tag is always zero now
        return speech_pb2.WordInfo(start_time=json_word['start'], end_time=json_word['end'],
                                   word=json_word['word'], confidence=json_word['conf'])

    def __process_partial_words(self, partial_words):

        def calculate_stability(words):
            stability = 0
            for word in words:
                stability += 0.02 + 0.002 * len(word)
            return stability / len(words)

        if len(partial_words) <= self.lm_order:
            return [' '.join(partial_words), calculate_stability(partial_words)],

        part1 = partial_words[:-self.lm_order]
        part2 = partial_words[-self.lm_order:]
        return [' '.join(part1), 0.9], [' '.join(part2), calculate_stability(part2)]

    def process_partial_result(self, json_partial_result, event_type):
        result = json.loads(json_partial_result)
        partial_result = result.get('partial', '')
        if len(partial_result) == 0 or self.last_partial_result == partial_result:
            return None

        self.last_partial_result = partial_result
        results = []
        for text, stability in self.__process_partial_words(partial_result.split(' ')):
            alternatives = [speech_pb2.SpeechRecognitionAlternative(transcript=text)]
            results.append(speech_pb2.StreamingRecognitionResult(alternatives=alternatives,
                                                                 stability=stability, channel_tag=1))

        return speech_pb2.StreamingRecognizeResponse(results=results, speech_event_type=event_type)

    def process_result(self, json_result, event_type):
        """
        results { alternatives { transcript: "to be or not to be" confidence: 0.92 }
                  alternatives { transcript: "to bee or not to bee" }
                  is_final: true }
        :param json_result:
        :param event_type:
        :return:
        """
        self.last_partial_result = ''
        result = json.loads(json_result)
        text = result['text'].strip()
        if len(text) == 0:
            return None

        words = [self.get_word_info(json_word) for json_word in result.get('result', [])]
        confidences = [word.confidence for word in words]
        first_alternative_confidence = sum(confidences) / len(confidences)
        # only the first alternative should has confidence
        alternatives = [speech_pb2.SpeechRecognitionAlternative(transcript=text, words=words,
                                                                confidence=first_alternative_confidence)]

        results = [speech_pb2.StreamingRecognitionResult(alternatives=alternatives, is_final=True, channel_tag=1)]

        return speech_pb2.StreamingRecognizeResponse(results=results, speech_event_type=event_type)

    def add_result(self, json_result, event_type):
        assert self.non_streaming
        result = json.loads(json_result)
        text = result['text'].strip()
        if len(text) == 0:
            return

        words = [self.get_word_info(json_word) for json_word in result.get('result', [])]
        confidences = [word.confidence for word in words]
        first_alternative_confidence = sum(confidences) / len(confidences)
        # only the first alternative should has confidence
        alternatives = [speech_pb2.SpeechRecognitionAlternative(transcript=text, words=words,
                                                                confidence=first_alternative_confidence)]

        self.results_buffer.append(speech_pb2.SpeechRecognitionResult(alternatives=alternatives,
                                                                      speech_event_type=event_type,
                                                                      channel_tag=1))

    def generate_recognize_response(self):
        assert self.non_streaming
        if len(self.results_buffer) == 0:
            return speech_pb2.RecognizeResponse()

        return speech_pb2.RecognizeResponse(results=self.results_buffer)


class DeepMineSpeechServicer(speech_pb2_grpc.SpeechServicer):
    """Provides methods that implement functionality of route guide server."""

    def __init__(self):
        self.supported_languages = set()
        self.supported_sample_rates = {8000, 11025, 16000, 22050, 32000, 44100, 48000}
        self.supported_models = defaultdict(set)
        self.supported_encodings = {AudioEncoding.ENCODING_UNSPECIFIED, AudioEncoding.LINEAR16}
        self.models = defaultdict(dict)
        self.thread_pool = futures.ThreadPoolExecutor(max_workers=10, thread_name_prefix='conv-')
        models = {}
        for lang, lang_models in AVAILABLE_MODELS.items():
            for model_name, (directory, frequency) in lang_models.items():
                hclg_path = os.path.join(directory, 'graph/HCLG.fst')
                hcl_path = os.path.join(directory, 'graph/HCLr.fst')
                if os.path.exists(hclg_path) or os.path.exists(hcl_path):
                    self.supported_languages.add(lang)
                    self.supported_models[lang].add(model_name)
                    if directory not in models:
                        print('Loading model from {}'.format(directory))
                        models[directory] = Model(directory)
                    self.models[lang][model_name] = (models[directory], frequency)

        self.segment_length_ms = 180

    @classmethod
    def check_user_quota(cls, context, nonbasic):
        metadata = dict(context.invocation_metadata())
        client_ip = metadata.get('client-ip')
        access_token = metadata.get('access-token')
        is_admin_call = False
        user_id = None
        if access_token is None:
            api_admin_token = metadata.get('admin-token')
            if api_admin_token:
                is_admin_call = True
                if api_admin_token != API_ADMIN_TOKEN:
                    logger.warning(f'Invalid admin token , client-ip: {client_ip}')
                    context.abort(grpc.StatusCode.UNAUTHENTICATED, InvalidAdminToken.detail())
                user_id = metadata.get('user-id')
                if not user_id:
                    logger.warning(f'No user id in admin call , client-ip: {client_ip}')
                    context.abort(grpc.StatusCode.UNAUTHENTICATED, NoUserId.detail())
                nonbasic = True  # in this case we should use nonbasic quota only.
            else:
                logger.warning(f'No access token , client-ip: {client_ip}')
                context.abort(grpc.StatusCode.UNAUTHENTICATED, NoAccessToken.detail())

        try:
            user_quota = QuotaManager.get_user_quota_class(access_token=access_token, user_id=user_id)
            user_quota.update_remaining_quota(access_token=access_token, user_id=user_id)
        except GrpcException as exp:
            logger.error(f'Exception in getting user_quota, exp: {str(exp)} , client-ip: {client_ip}')
            context.abort(grpc.StatusCode.UNAUTHENTICATED, exp.detail())
            return

        # if remaining quota is less than MIN_QUOTA_SEC seconds
        if user_quota.get_remaining_quota(nonbasic) < MIN_QUOTA_SEC:
            logger.warning(f'Remaining quota less than MIN_QUOTA_SEC , client-ip: {client_ip}')
            if nonbasic:
                context.abort(grpc.StatusCode.RESOURCE_EXHAUSTED, NoRemainingNonbasicQuotaError.detail())
            else:
                context.abort(grpc.StatusCode.RESOURCE_EXHAUSTED, NoRemainingQuotaError.detail())

        return user_quota, nonbasic, is_admin_call, client_ip

    def check_config(self, config, context, is_streaming=False):
        if len(config.model) == 0:
            config.model = 'dictation'
        if not config.encoding:
            config.encoding = AudioEncoding.ENCODING_UNSPECIFIED
        if config.max_alternatives <= 1:
            config.max_alternatives = 0
        if len(config.language_code) == 0:
            details = InvalidArgumentError.detail(InvalidArgumentError.default_detail +
                                                  ' The language code should be specified.')
            context.abort(code=grpc.StatusCode.INVALID_ARGUMENT, details=details)
        if config.language_code not in self.supported_languages:
            details = UnimplementedError.detail(UnimplementedError.default_detail +
                                                f' The language code `{config.language_code}` is not supported yet.')
            context.abort(code=grpc.StatusCode.UNIMPLEMENTED, details=details)
        if config.model not in self.supported_models[config.language_code]:
            details = UnimplementedError.detail(UnimplementedError.default_detail +
                                                f' The model `{config.model}` is not supported for the'
                                                f' language `{config.language_code}`.')
            context.abort(code=grpc.StatusCode.UNIMPLEMENTED, details=details)
        if config.sample_rate_hertz not in self.supported_sample_rates:
            details = UnimplementedError.detail(UnimplementedError.default_detail +
                                                f' The sampling rate `{config.sample_rate_hertz}` is not supported yet.'
                                                f' We only support `{self.supported_sample_rates}` rates.')
            context.abort(code=grpc.StatusCode.UNIMPLEMENTED, details=details)
        if is_streaming and config.encoding == AudioEncoding.ENCODING_UNSPECIFIED:
            details = InvalidArgumentError.detail(InvalidArgumentError.default_detail +
                                                  ' For streaming mode, the audio encoding should be specified.')
            context.abort(code=grpc.StatusCode.INVALID_ARGUMENT, details=details)
        if config.max_alternatives > 1:
            details = UnimplementedError.detail(UnimplementedError.default_detail +
                                                ' Max alternatives grater than one is not supported yet.')
            context.abort(code=grpc.StatusCode.UNIMPLEMENTED, details=details)
        if config.enable_speaker_diarization:
            details = UnimplementedError.detail(UnimplementedError.default_detail +
                                                ' Speaker diarization is not supported yet. Set `enable_speaker_'
                                                'diarization` to false to disable it.')
            context.abort(code=grpc.StatusCode.UNIMPLEMENTED, details=details)
        if config.enable_automatic_punctuation:
            details = UnimplementedError.detail(UnimplementedError.default_detail +
                                                ' Automatic punctuation prediction is not supported yet. Set '
                                                '`enable_automatic_punctuation` to false to disable it.')
            context.abort(code=grpc.StatusCode.UNIMPLEMENTED, details=details)
        if config.enable_separate_recognition_per_channel:
            details = UnimplementedError.detail(UnimplementedError.default_detail +
                                                ' Separate recognition per channel is not supported yet. Set '
                                                '`enable_separate_recognition_per_channel` to false to disable it.')
            context.abort(code=grpc.StatusCode.UNIMPLEMENTED, details=details)
        if config.profanity_filter:
            details = UnimplementedError.detail(UnimplementedError.default_detail +
                                                ' Automatic profanity filtering is not supported yet. Set '
                                                '`profanity_filter` to false to disable it.')
            context.abort(code=grpc.StatusCode.UNIMPLEMENTED, details=details)

    @staticmethod
    def download_uri(context, uri):
        if not validators.url(uri):
            logger.warning(f'Invalid URI: {uri}')
            context.abort(code=grpc.StatusCode.INVALID_ARGUMENT,
                          details=InvalidUriFormatError.detail(detail=InvalidUriFormatError.default_detail + uri))
        content, err = download_file(uri)
        if not content:
            logger.warning(f'Unable to get content from URI: {uri}')
            context.abort(code=grpc.StatusCode.INVALID_ARGUMENT,
                          details=UriDownloadError.detail(detail=UriDownloadError.default_detail + err))
        return content

    def Recognize(self, request, context):
        nonbasic = False
        user_quota, nonbasic, is_admin_call, client_ip = self.check_user_quota(context, nonbasic=nonbasic)

        def log_message(_message, level=logging.INFO, print_console=False):
            logger.log(msg=f'uid: {user_quota.user_id:<5} ip: {client_ip:<15} {_message}', level=level)
            if print_console:
                print(f'uid: {user_quota.user_id:<5} ip: {client_ip:<15} {_message}')

        log_message(f'Recognize, is_admin_call: {is_admin_call}', print_console=True)

        config = request.config
        audio = request.audio
        if audio.HasField("content"):
            uri = None
            content = audio.content
        else:
            uri = audio.uri
            if IS_DEVELOPMENT or is_admin_call:
                log_message(f'URI: {uri}')
            content = None

        self.check_config(config, context)
        if content is None:
            content = DeepMineSpeechServicer.download_uri(context, uri)

        model = config.model
        language_code = config.language_code
        selected_model, sample_rate_hertz = self.models[language_code][model]
        input_size = len(content)
        converter = 'unknown'
        try:
            if config.encoding == AudioEncoding.ENCODING_UNSPECIFIED:
                converter = 'ffmpeg'
                content = Converter.audio_buffer_to_pipe(content, sample_rate_hertz)
            else:
                if sample_rate_hertz == config.sample_rate_hertz and config.encoding == AudioEncoding.LINEAR16:
                    if IS_DEVELOPMENT or is_admin_call:
                        log_message('Pass format conversion.')
                    if SAVE_AUDIOS and False:
                        name = ''.join(random.choices(string.ascii_uppercase + string.digits, k=8))
                        with open('files/' + name + '.wav', 'wb') as fid2:
                            fid2.write(content)
                else:
                    converter = 'miniaudio'
                    content = Converter.known_audio_buffer_to_pipe(content, sample_rate_hertz)
        except GrpcException as exp:
            log_message(f'Exception in converting file, [{exp.detail()}]', level=logging.ERROR)
            context.abort(code=grpc.StatusCode.INVALID_ARGUMENT, details=exp.detail())

        if len(content) == 0:
            log_message(f'Empty content.', level=logging.WARN)
            context.abort(code=grpc.StatusCode.INVALID_ARGUMENT, details=AudioContentError.detail())

        recognizer = KaldiRecognizer(selected_model, sample_rate_hertz)
        response_generator = ResponseGenerator(non_streaming=True)
        recognizer.SetMaxAlternatives(0)
        recognizer.SetWords(True)

        duration = 0
        start_time = time.time()
        continue_recognition = True
        start_idx, _len = 0, len(content)
        byte_rate_sec = sample_rate_hertz * 2.0
        content_duration = _len / byte_rate_sec
        segment_length = int(sample_rate_hertz * self.segment_length_ms * 2 / 1000)

        if converter == 'unknown':
            log_message(f'Unconverted audio has {input_size} bytes length and {content_duration:.2f} seconds duration.')
        else:
            log_message(f'Converted audio with len {input_size} to len {_len} using {converter}. '
                        f'Duration is {content_duration:.2f} sec')

        quota = user_quota.get_remaining_quota(nonbasic)
        if quota < content_duration:
            err = f' Remaining quota is {quota} seconds while the audio duration is {content_duration:.2f} seconds.'
            log_message(f'Not enough remaining quota for, {err}', level=logging.WARN)
            if nonbasic:
                details = NoRemainingNonbasicQuotaError.detail(NoRemainingNonbasicQuotaError.default_detail + err)
                context.abort(code=grpc.StatusCode.RESOURCE_EXHAUSTED, details=details)
            else:
                details = NoRemainingQuotaError.detail(NoRemainingQuotaError.default_detail + err)
                context.abort(code=grpc.StatusCode.RESOURCE_EXHAUSTED, details=details)

        # def done():
        #     print('Terminated')
        #
        # context.add_callback(done)

        try:
            if is_admin_call:
                log_message('Start processing loop ...')
            event_type = SpeechEventType.END_OF_SPEECH
            while start_idx < _len and continue_recognition:
                if not context.is_active():
                    event_type = SpeechEventType.CANCELLED
                    log_message("Call was cancelled by the client.")
                    continue_recognition = False
                    break
                segment = content[start_idx: start_idx + segment_length]
                chunk_duration = len(segment) / byte_rate_sec
                duration += chunk_duration
                result = recognizer.AcceptWaveform(segment)
                remaining_quota = user_quota.use_quota(chunk_duration, nonbasic=False, client_ip=client_ip,
                                                       is_admin_call=is_admin_call)
                if remaining_quota < MIN_QUOTA_SEC:
                    continue_recognition = False
                    event_type = SpeechEventType.QUOTA_FINISHED
                    response_generator.add_result(recognizer.FinalResult(), event_type)

                elif result:
                    if time.time() - start_time > TIMEOUT_SEC:
                        continue_recognition = False
                        event_type = SpeechEventType.TIMEOUT
                        response_generator.add_result(recognizer.FinalResult(), event_type)
                    elif duration > DURATION_LIMIT_SEC:
                        continue_recognition = False
                        event_type = SpeechEventType.DATA_SIZE_LIMIT
                        response_generator.add_result(recognizer.FinalResult(), event_type)
                    else:
                        if IS_DEVELOPMENT and is_admin_call:
                            log_message(f'END_POINT @ start_idx: {start_idx}')
                        event_type = SpeechEventType.END_POINT
                        response_generator.add_result(recognizer.Result(), event_type)

                start_idx += segment_length

            if continue_recognition:
                event_type = SpeechEventType.END_OF_SPEECH
                response_generator.add_result(recognizer.FinalResult(), event_type)
        finally:
            # roundup second used to 10 seconds
            rem = duration % 10
            if rem >= 1:
                # print('Roundup usage with %.2f seconds' % (10 - rem))
                user_quota.use_quota(10 - rem, nonbasic=False, client_ip=client_ip,
                                     for_roundup=True, is_admin_call=is_admin_call)
            user_quota.finalize_usage(client_ip)

        uri_uuid = '-'
        if uri:
            try:
                uri_uuid = ', uuid: ' + uri.split('/')[-1]
            except IndexError:
                pass
        log_message(f"Recognize finished with event_type: {event_type.name} for {uri_uuid}")

        response = response_generator.generate_recognize_response()
        return response

    @staticmethod
    def _get_content(context, request_iterator, streaming_content: StreamingContent):
        print('Starting get_content')
        while True:
            request = next(request_iterator, None)
            if not request:
                break
            if not request.HasField("audio_content"):
                # Error in the request, all subsequent messages must only have audio_content.
                context.abort(code=grpc.StatusCode.INVALID_ARGUMENT,
                              details=InvalidSubsequentStreamingRequest.detail())
            streaming_content.write(request.audio_content)
            print(f'Received content with len {len(request.audio_content)}')
        streaming_content.close()

    @staticmethod
    def _simple_content_iterator(context, request_iterator, user_id, client_ip):
        # print('Starting get_content')
        while True:
            request = next(request_iterator, None)
            if not request:
                break
            if not request.HasField("audio_content"):
                # Error in the request, all subsequent messages must only have audio_content.
                logger.error(f'uid: {user_id:>5} ip: {client_ip:>15} Exception in reading stream content, '
                             f'{InvalidSubsequentStreamingRequest.detail()}')
                context.abort(code=grpc.StatusCode.INVALID_ARGUMENT,
                              details=InvalidSubsequentStreamingRequest.detail())
            yield request.audio_content
            # print(f'Received content with len {len(request.audio_content)}')

    def StreamingRecognize(self, request_iterator, context):
        print('Start StreamingRecognize')

        user_quota, _, _, client_ip = self.check_user_quota(context, nonbasic=True)

        request = next(request_iterator)
        if not request.HasField("streaming_config"):
            # Error in the request, the first one must only have streaming_config.
            context.abort(code=grpc.StatusCode.INVALID_ARGUMENT,
                          details=InvalidFirstStreamingRequest.detail())

        config = request.streaming_config.config
        self.check_config(config, context, is_streaming=True)

        model = config.model
        encoding = config.encoding
        language_code = config.language_code
        max_alternatives = config.max_alternatives
        enable_word_confidence = config.enable_word_confidence
        enable_word_time_offsets = config.enable_word_time_offsets
        single_utterance = request.streaming_config.single_utterance
        selected_model, sample_rate_hertz = self.models[language_code][model]

        recognizer = KaldiRecognizer(selected_model, sample_rate_hertz)
        recognizer.SetMaxAlternatives(max_alternatives)
        recognizer.SetWords(True)
        response_generator = ResponseGenerator(non_streaming=True, enable_word_confidence=enable_word_confidence,
                                               enable_word_time_offsets=enable_word_time_offsets)

        duration = 0
        start_time = time.time()
        continue_recognition = True
        byte_rate_sec = sample_rate_hertz * 2.0
        if SAVE_AUDIOS and False:
            file_name = ''.join(random.choices(string.ascii_uppercase + string.digits, k=8))
            out_file = open('files/live_' + file_name + '.wav', 'wb')
        else:
            out_file = None

        if sample_rate_hertz == config.sample_rate_hertz and encoding == AudioEncoding.LINEAR16:
            segment_iterator = self._simple_content_iterator(context, request_iterator, user_quota.user_id, client_ip)
        else:
            streaming_content = StreamingContent()
            # future = self.thread_pool.submit(self._get_content, (context, request_iterator, streaming_content))
            # print(future.done())
            thread = threading.Thread(target=self._get_content, args=(context, request_iterator, streaming_content))
            thread.daemon = True
            thread.start()
            try:
                segment_iterator = Converter.conversion_stream(streaming_content, encoding, sample_rate_hertz)
            except miniaudio.DecodeError:
                context.abort(code=grpc.StatusCode.FAILED_PRECONDITION,
                              details=InvalidStreamHeader.detail())
                return
        try:
            while continue_recognition:
                # print('Waiting for the next segment ...')
                segment = next(segment_iterator, None)
                if segment is None:
                    break
                chunk_duration = len(segment) / byte_rate_sec
                duration += chunk_duration
                if isinstance(segment, array.array):
                    segment = segment.tobytes()
                if SAVE_AUDIOS:
                    out_file.write(segment)

                result = recognizer.AcceptWaveform(segment)
                remaining_quota = user_quota.use_quota(chunk_duration, nonbasic=True, client_ip=client_ip)

                if remaining_quota < MIN_QUOTA_SEC:
                    continue_recognition = False
                    event_type = SpeechEventType.QUOTA_FINISHED
                    response_generator.add_result(recognizer.FinalResult(), event_type)
                elif result:
                    if time.time() - start_time > TIMEOUT_SEC:
                        continue_recognition = False
                        event_type = SpeechEventType.TIMEOUT
                        response_generator.add_result(recognizer.FinalResult(), event_type)
                    elif duration > DURATION_LIMIT_SEC:
                        continue_recognition = False
                        event_type = SpeechEventType.DATA_SIZE_LIMIT
                        response_generator.add_result(recognizer.FinalResult(), event_type)
                    elif single_utterance:
                        continue_recognition = False
                        event_type = SpeechEventType.END_OF_SINGLE_UTTERANCE
                        response_generator.add_result(recognizer.FinalResult(), event_type)
                    else:
                        event_type = SpeechEventType.END_POINT
                        response_generator.add_result(recognizer.Result(), event_type)

            if continue_recognition:
                event_type = SpeechEventType.END_OF_SPEECH
                response_generator.add_result(recognizer.FinalResult(), event_type)

            if SAVE_AUDIOS:
                print('Request finished. Close the file.')
                out_file.close()
        finally:
            # roundup second used to 10 seconds
            rem = duration % 10
            if rem >= 1:
                # print('Roundup usage with %.2f seconds' % (10 - rem))
                user_quota.use_quota(10 - rem, nonbasic=True, client_ip=client_ip, for_roundup=True)
            user_quota.finalize_usage(client_ip)

        return response_generator.generate_recognize_response()

    def BidiStreamingRecognize(self, request_iterator, context):
        user_quota, _, _, client_ip = self.check_user_quota(context, nonbasic=True)

        def log_message(_message, level=logging.INFO, print_console=False):
            logger.log(msg=f'uid: {user_quota.user_id:<5} ip: {client_ip:<15} {_message}', level=level)
            if print_console:
                print(f'uid: {user_quota.user_id:<5} ip: {client_ip:<15} {_message}')

        log_message('BidiStreamingRecognize', print_console=True)

        request = next(request_iterator)
        if not request.HasField("streaming_config"):
            # Error in the request, the first one must only have streaming_config.
            log_message(InvalidFirstStreamingRequest.detail(), level=logging.ERROR)
            context.abort(code=grpc.StatusCode.INVALID_ARGUMENT,
                          details=InvalidFirstStreamingRequest.detail())

        config = request.streaming_config.config
        self.check_config(config, context, is_streaming=True)

        model = config.model
        encoding = config.encoding
        language_code = config.language_code
        max_alternatives = config.max_alternatives
        enable_word_confidence = config.enable_word_confidence
        enable_word_time_offsets = config.enable_word_time_offsets
        interim_results = request.streaming_config.interim_results
        single_utterance = request.streaming_config.single_utterance
        # interim_results_interval_milliseconds = request.streaming_config.interim_results_interval_milliseconds
        selected_model, sample_rate_hertz = self.models[language_code][model]

        recognizer = KaldiRecognizer(selected_model, sample_rate_hertz)
        recognizer.SetMaxAlternatives(max_alternatives)
        recognizer.SetWords(True)
        response_generator = ResponseGenerator(non_streaming=False, enable_word_confidence=enable_word_confidence,
                                               enable_word_time_offsets=enable_word_time_offsets)

        duration = 0
        start_time = time.time()
        continue_recognition = True
        byte_rate_sec = sample_rate_hertz * 2.0

        if sample_rate_hertz == config.sample_rate_hertz and encoding == AudioEncoding.LINEAR16:
            segment_iterator = self._simple_content_iterator(context, request_iterator, user_quota.user_id, client_ip)
        else:
            streaming_content = StreamingContent()
            thread = threading.Thread(target=self._get_content, args=(context, request_iterator, streaming_content))
            thread.daemon = True
            thread.start()
            try:
                segment_iterator = Converter.conversion_stream(streaming_content, encoding, sample_rate_hertz)
            except miniaudio.DecodeError as exp:
                log_message(f'Exception in converting stream, [{str(exp)}]', level=logging.ERROR)
                context.abort(code=grpc.StatusCode.FAILED_PRECONDITION,
                              details=InvalidStreamHeader.detail())
                return

        # open a file if we should save the input speech
        if SAVE_AUDIOS:
            dir_name = f'files/{user_quota.user_id:0>6}'
            if not os.path.exists(dir_name):
                try:
                    os.makedirs(dir_name)
                except IOError:
                    pass
            random_id = "".join(random.choices(string.ascii_letters + string.digits, k=8))
            file_path = f'{dir_name}/{time.strftime("%Y-%m-%d_%H-%M-%S")}_{random_id}.wav'
            log_message(f'Save Bidi speech in {file_path}')
            out_file = open(file_path, 'wb', buffering=32768)
        else:
            out_file = None

        try:
            event_type = SpeechEventType.END_OF_SPEECH
            while continue_recognition:
                segment = next(segment_iterator, None)
                if segment is None:
                    break
                chunk_duration = len(segment) / byte_rate_sec
                duration += chunk_duration
                if isinstance(segment, array.array):
                    segment = segment.tobytes()
                if SAVE_AUDIOS:
                    out_file.write(segment)
                result = recognizer.AcceptWaveform(segment)
                remaining_quota = user_quota.use_quota(chunk_duration, client_ip=client_ip, nonbasic=True)

                if remaining_quota < MIN_QUOTA_SEC:
                    continue_recognition = False
                    event_type = SpeechEventType.QUOTA_FINISHED
                    response = response_generator.process_result(recognizer.FinalResult(), event_type)
                    if response:
                        yield response

                elif result:
                    if time.time() - start_time > TIMEOUT_SEC:
                        continue_recognition = False
                        event_type = SpeechEventType.TIMEOUT
                    elif duration > DURATION_LIMIT_SEC:
                        continue_recognition = False
                        event_type = SpeechEventType.DATA_SIZE_LIMIT
                    elif single_utterance:
                        continue_recognition = False
                        event_type = SpeechEventType.END_OF_SINGLE_UTTERANCE
                    else:
                        event_type = SpeechEventType.END_POINT
                    response = response_generator.process_result(recognizer.Result(), event_type)
                    if response:
                        yield response

                elif interim_results:
                    if time.time() - start_time > TIMEOUT_SEC:
                        event_type = SpeechEventType.TIMEOUT
                        continue_recognition = False
                    elif duration > DURATION_LIMIT_SEC:
                        event_type = SpeechEventType.DATA_SIZE_LIMIT
                        continue_recognition = False
                    else:
                        event_type = SpeechEventType.SPEECH_EVENT_UNSPECIFIED
                    response = response_generator.process_partial_result(recognizer.PartialResult(), event_type)
                    if response:
                        yield response

            if continue_recognition:
                event_type = SpeechEventType.END_OF_SPEECH
                response = response_generator.process_result(recognizer.FinalResult(), event_type)
                if response:
                    yield response
        finally:
            if SAVE_AUDIOS:
                log_message('Request finished. Close the file.')
                out_file.close()
            # roundup second used to 10 seconds
            rem = duration % 10
            if rem >= 1:
                # log_message('Roundup usage with %.2f seconds' % (10 - rem))
                user_quota.use_quota(10 - rem, nonbasic=True, client_ip=client_ip, for_roundup=True)
            user_quota.finalize_usage(client_ip)
        log_message(f"BidiStreamingRecognize finished with event_type: {event_type.name}")

    def Align(self, request, context):
        print('Start Align')

        config = request.config
        audio = request.audio
        text = request.text

        if audio.HasField("content"):
            content = audio.content
        else:
            uri = audio.uri
            content = ''

        self.check_config(config, context)
        sample_rate_hertz = config.sample_rate_hertz

        aligner = Aligner(text=text)
        recognizer = KaldiRecognizer(self.model, sample_rate_hertz)
        recognizer.SetMaxAlternatives(0)
        recognizer.SetWords(True)

        segment_length = int(sample_rate_hertz * self.segment_length_ms * 2 / 1000)
        start_idx, _len = 0, len(content)
        while start_idx < _len:
            segment = content[start_idx: start_idx + segment_length]
            result = recognizer.AcceptWaveform(segment)
            if result:
                aligner.add_result(recognizer.Result())
            start_idx += segment_length

        aligner.add_result(recognizer.FinalResult())
        response = aligner.do_alignment()

        return response


def serve():
    with open('cert/server.key', 'rb') as fid:
        private_key = fid.read()

    with open('cert/server.crt', 'rb') as fid:
        certificate_chain = fid.read()

    # https://www.sandtable.com/using-ssl-with-grpc-in-python/
    server_credentials = grpc.ssl_server_credentials(((private_key, certificate_chain),))

    server = grpc.server(futures.ThreadPoolExecutor(max_workers=MAX_WORKERS),
                         maximum_concurrent_rpcs=MAXIMUM_CONCURRENT_RPCS,
                         options=[
                             ('grpc.max_send_message_length', MAX_MESSAGE_LENGTH),
                             ('grpc.max_receive_message_length', MAX_MESSAGE_LENGTH),
                             ('grpc.keepalive_time_ms', 30000),
                             # send keepalive ping every 30 second, default is 2 hours
                             ('grpc.keepalive_timeout_ms', 15000),
                             # keepalive ping time out after 15 seconds, default is 20 seconds
                             ('grpc.keepalive_permit_without_calls', True),
                             # allow keepalive pings when there's no gRPC calls
                             ('grpc.http2.max_pings_without_data', 0),
                             # allow unlimited amount of keepalive pings without data
                             ('grpc.http2.min_time_between_pings_ms', 15000),
                             # allow grpc pings from client every 15 seconds
                             ('grpc.http2.min_ping_interval_without_data_ms', 10000),
                             # allow grpc pings from client without data every 10 seconds
                         ])

    speech_pb2_grpc.add_SpeechServicer_to_server(DeepMineSpeechServicer(), server)

    server.add_secure_port('{}:{}'.format(ASR_SERVER_INTERFACE, ASR_SERVER_PORT), server_credentials)
    server.start()
    server.wait_for_termination()


if __name__ == '__main__':
    logging.basicConfig()
    serve()
